@@ -7,16 +7,13 @@ email programmer@usa.com
 # Path
 import os
 import sys
-import importlib
-import inspect
-import tempfile
 from pathlib import Path
 import string
 import random
 
 # Numerical
 import numpy as np
-import pandas as pd
+# import pandas as pd  # heavy weight, avoid unless necessary.
 # Meta
 import dill
 import copy
@@ -26,6 +23,9 @@ import datetime as dt  # useful for deltatime and timezones.
 _ = dt
 
 
+# ============================== Accessories ============================================
+
+
 def timestamp(fmt=None, name=None):
     """isoformat is not compatible with file naming convention, this function provides compatible fmt
     tip: do not use this to create random addresses as it fails at high speed runs. Random string is better."""
@@ -54,7 +54,10 @@ def str2timedelta(past):
     return dt.timedelta(**{key: val})
 
 
-def randstr(length=10, lower=True, upper=True, digits=True, punctuation=False):
+def randstr(length=10, lower=True, upper=True, digits=True, punctuation=False, safe=False):
+    if safe:
+        import secrets  # interannly, it uses: random.SystemRandom or os.urandom which is hardware-based, not pseudo
+        return secrets.token_urlsafe(length)
     pool = "" + (string.ascii_lowercase if lower else "")
     pool = pool + (string.ascii_uppercase if upper else "")
     pool = pool + (string.digits if digits else "")
@@ -63,6 +66,21 @@ def randstr(length=10, lower=True, upper=True, digits=True, punctuation=False):
     return result_str
 
 
+def assert_package_installed(package):
+    """imports a package and installs it if not."""
+    try:
+        pkg = __import__(package)
+        return pkg
+    except ImportError:
+        import subprocess
+        subprocess.check_call([sys.executable, "-m", "pip", "install", package])
+    pkg = __import__(package)
+    return pkg
+
+
+# ====================================== Classes ====================================
+
+
 class SaveDecorator(object):
     def __init__(self, func, ext=""):
         # TODO: migrate from save_decorator to SaveDecorator
@@ -85,6 +103,7 @@ class SaveDecorator(object):
     def __call__(self, path=None, obj=None, **kwargs):
         # Called when calling the decorated function (instance of this called).
         if path is None:
+            import tempfile
             path = Path(tempfile.mkdtemp() + "-" + timestamp() + self.ext)
             # raise ValueError
         else:
@@ -124,8 +143,7 @@ def save_decorator(ext=""):
             if verbose:
                 rep = repr(obj)
                 rep = rep if len(rep) < 50 else rep[:10] + "... "
-                print(f"Object {rep} saved @ ", path.absolute().as_uri(),
-                      " |  Directory: ", path.parent.absolute().as_uri())
+                print(f"SAVED {rep}  @ `{path.absolute().as_uri()}` |  Directory: `{path.parent.absolute().as_uri()}`")
             return path
 
         return wrapper
@@ -227,6 +245,7 @@ class Base(object):
     def save_code(self, path):
         """a usecase for including code in the save is when the source code is continously
          changing and still you want to reload an old version."""
+        import inspect
         module = inspect.getmodule(self)
         if hasattr(module, "__file__"): file = Path(module.__file__)
         else: raise FileNotFoundError(f"Attempted to save code from a script running in interactive session! "
@@ -340,6 +359,7 @@ class Base(object):
         """
         code_path = Path(code_path)
         sys.path.insert(0, str(code_path.parent))
+        import importlib
         sourcefile = importlib.import_module(code_path.stem)
         return getattr(sourcefile, class_name).from_state(data_path, *args, r=r, **kwargs)
 
@@ -382,6 +402,7 @@ class Base(object):
             [attrs.remove(x) for x in Base().get_attributes(remove_base_attrs=False)]
         # if exclude is not None:
         #     [attrs.remove(x) for x in exlcude]
+        import inspect
         if not fields:  # logic (questionable): anything that is not a method is a field
             attrs = list(filter(lambda x: inspect.ismethod(getattr(self, x)), attrs))
         
@@ -426,7 +447,7 @@ class Base(object):
             return string_
 
     def print(self, typeinfo=False):
-        Struct(self.__dict__).print(typeinfo=typeinfo)
+        Struct(self.__dict__).print(dtype=typeinfo)
 
     def viz_heirarchy(self, depth=3, obj=None, filt=None):
         import objgraph
@@ -673,9 +694,7 @@ class List(list, Base):
             self.apply(lambda x: x.apply(func, *args, other=other, jobs=jobs, depth=depth, **kwargs))
 
         func = self.evalstr(func, expected='func')
-        # tqdm = 0
-        # Experimental.assert_package_installed("tqdm")  # removed for disentanglement
-        from tqdm import tqdm
+        tqdm = assert_package_installed("tqdm").tqdm
         if other is None:
             iterator = self.list if not verbose else tqdm(self.list, desc=desc)
             if jobs:
@@ -731,6 +750,10 @@ class List(list, Base):
             if sep:
                 print(sep * 100)
 
+    def to_series(self):
+        import pandas as pd
+        return pd.Series(self.list)
+
     def to_dataframe(self, names=None, minimal=False, obj_included=True):
         """
 
@@ -743,6 +766,7 @@ class List(list, Base):
         columns = list(self.list[0].__dict__.keys())
         if obj_included or names:
             columns = ['object'] + columns
+        import pandas as pd
         df = pd.DataFrame(columns=columns)
         if minimal:
             return df
@@ -767,10 +791,18 @@ class List(list, Base):
         return np.array(self.list)
 
 
-class Struct(Base):  # inheriting from dict gives `get` method.
+class Struct(dict):  # inheriting from dict gives `get` method.
     """Use this class to keep bits and sundry items.
     Combines the power of dot notation in classes with strings in dictionaries to provide Pandas-like experience
     """
+
+    def save_json(self, path=None):
+        path = Save.json(obj=self.__dict__, path=path)
+        return path
+
+    def save_yaml(self, path=None):
+        return Save.yaml(obj=self.__dict__, path=path)
+
     def __len__(self):
         return len(self.keys())
 
@@ -860,7 +892,9 @@ class Struct(Base):  # inheriting from dict gives `get` method.
             repr_string += str(key) + ", "
         return "Struct: [" + repr_string + "]"
 
-    def print(self, sep=None, yaml=False, typeinfo=True, logger=False):
+    def print(self, sep=None, yaml=False, dtype=True, logger=False, limit=50, config=False, newline=True):
+        if config:
+            return print(DisplayData.config(self.__dict__, newline=newline))
         if bool(self) is False:
             print(f"Empty Struct.")
             return None  # break out of the function.
@@ -874,14 +908,14 @@ class Struct(Base):  # inheriting from dict gives `get` method.
             sep = 5 + max(self.keys().apply(str).apply(len).list)
         repr_string = ""
         repr_string += "Structure, with following entries:\n"
-        repr_string += "Key" + " " * sep + (("Item Type" + " " * sep) if typeinfo else "") + "Item Details\n"
-        repr_string += "---" + " " * sep + (("---------" + " " * sep) if typeinfo else "") + "------------\n"
+        repr_string += "Key" + " " * sep + (("Item Type" + " " * sep) if dtype else "") + "Item Details\n"
+        repr_string += "---" + " " * sep + (("---------" + " " * sep) if dtype else "") + "------------\n"
         for key in self.keys().list:
             key_str = str(key)
             type_str = str(type(self[key])).split("'")[1]
-            val_str = DisplayData.get_repr(self[key]).replace("\n", " ")
+            val_str = DisplayData.get_repr(self[key], limit=limit).replace("\n", " ")
             repr_string += key_str + " " * abs(sep - len(key_str)) + " " * len("Key")
-            if typeinfo:
+            if dtype:
                 repr_string += type_str + " " * abs(sep - len(type_str)) + " " * len("Item Type")
             repr_string += val_str + "\n"
         if logger:
@@ -1003,6 +1037,7 @@ class Struct(Base):  # inheriting from dict gives `get` method.
 
     def to_dataframe(self, *args, **kwargs):
         # return self.values().to_dataframe(names=self.keys())
+        import pandas as pd
         return pd.DataFrame(self.__dict__, *args, **kwargs)
 
     def spawn_from_values(self, values):
@@ -1031,11 +1066,11 @@ class Struct(Base):  # inheriting from dict gives `get` method.
 
 
 class DisplayData:
-    def __init__(self, x):
-        self.x = pd.DataFrame(x)
+    # primitive = {list, dict, tuple}
 
     @staticmethod
     def set_pandas_display(rows=1000, columns=1000, width=1000, colwidth=40):
+        import pandas as pd
         pd.set_option('display.max_colwidth', colwidth)
         pd.set_option('display.max_columns', columns)  # to avoid replacing them with ...
         pd.set_option('display.width', width)  # to avoid wrapping the table.
@@ -1044,10 +1079,12 @@ class DisplayData:
     @staticmethod
     def set_pandas_auto_width():
         """For fixed width host windows, this is recommended to avoid chaos due to line-wrapping."""
+        import pandas as pd
         pd.options.display.width = 0  # this way, pandas is told to detect window length and act appropriately.
 
     @staticmethod
     def eng():
+        import pandas as pd
         pd.set_eng_float_format(accuracy=3, use_eng_prefix=True)
         pd.options.display.float_format = '{:, .5f}'.format
         pd.set_option('precision', 7)
@@ -1069,12 +1106,19 @@ class DisplayData:
         return string_
 
     @staticmethod
-    def outline(array, name="Array", imprint=True):
+    def outline(array, name="Array", printit=True):
         str_ = f"{name}. Shape={array.shape}. Dtype={array.dtype}"
-        if imprint:
-            print(str_)
+        if printit: print(str_)
         return str_
 
+    @staticmethod
+    def config(mydict, newline=True):
+        rep = ""  # returns json repr
+        for key, val in mydict.items():
+            rep += f"{key}={val}"
+            rep += "\n" if newline else ", "
+        return rep
+
     @staticmethod
     def print_string_list(mylist, char_per_row=125, sep=" "):
         counter = 0
