@@ -1,9 +1,8 @@
 
 import logging
-import dill
 import subprocess
 import time
-from crocodile.core import np, os, sys, inspect, importlib, timestamp, randstr, str2timedelta, datetime, pd
+from crocodile.core import np, os, sys, timestamp, randstr, str2timedelta, datetime, Save, assert_package_installed
 from crocodile.file_management import P
 
 
@@ -82,6 +81,14 @@ class DictCycle(Cycle):
 class Experimental:
     """Debugging and Meta programming tools"""
 
+    @staticmethod
+    def profile_memory(command):
+        psutil = assert_package_installed("psutil")
+        before = psutil.virtual_memory()
+        exec(command)
+        after = psutil.virtual_memory()
+        print(f"Memory used = {(after.used - before.used) / 1e6}")
+
     @staticmethod
     def try_this(func, otherwise=None):
         try:
@@ -98,17 +105,6 @@ class Experimental:
         res = res.filter(lambda x: x not in {"In", "Out", "get_ipython", "quit", "exit", "sys"})
         res.print(**kwargs)
 
-    @staticmethod
-    def assert_package_installed(package):
-        """imports a package and installs it if not."""
-        try:
-            pkg = __import__(package)
-            return pkg
-        except ImportError:
-            subprocess.check_call([sys.executable, "-m", "pip", "install", package])
-        pkg = __import__(package)
-        return pkg
-
     @staticmethod
     def generate_readme(path, obj=None, meta=None, save_source_code=True):
         """Generates a readme file to contextualize any binary files.
@@ -119,6 +115,7 @@ class Experimental:
         :param meta:
         :param save_source_code:
         """
+        import inspect
         path = P(path)
         readmepath = path / f"README.md" if path.is_dir() else path
 
@@ -134,11 +131,11 @@ class Experimental:
             text += f"# Source code file generated me was located here: \n'{inspect.getfile(obj)}'\n" + separator
 
         readmepath.write_text(text)
-        print(f"Successfully generated README.md file. Checkout:\n", readmepath.as_uri())
+        print(f"Successfully generated README.md file. Checkout:\n", readmepath.absolute().as_uri())
 
         if save_source_code:
             P(inspect.getmodule(obj).__file__).zip(op_path=readmepath.with_name("source_code.zip"))
-            print("Source code saved @ " + readmepath.with_name("source_code.zip").as_uri())
+            print("Source code saved @ " + readmepath.with_name("source_code.zip").absolute().as_uri())
 
     @staticmethod
     def load_from_source_code(directory, obj=None, delete=False):
@@ -152,7 +149,7 @@ class Experimental:
         P(directory).find("source_code*", r=True).unzip(tmpdir)
         sys.path.insert(0, str(tmpdir))
         sourcefile = __import__(tmpdir.find("*").stem)
-        tmpdir.delete(are_you_sure=delete, verbose=False)
+        tmpdir.delete(sure=delete, verbose=False)
         if obj is not None:
             loaded = getattr(sourcefile, obj)
             return loaded
@@ -235,7 +232,7 @@ class Experimental:
             code_string = args_kwargs + code_string
 
         if copy2clipboard:
-            clipboard = Experimental.assert_package_installed("clipboard")
+            clipboard = assert_package_installed("clipboard")
             clipboard.copy(code_string)
         if verbose: print(f"code to be run extracted from {func.__name__} \n", code_string, "=" * 100)
         return code_string  # ready to be run with exec()
@@ -250,7 +247,8 @@ class Experimental:
             _ = self
             func = eval(func, modules)
 
-        from file_management import Struct
+        from crocodile.file_management import Struct
+        import inspect
         ak = Struct(dict(inspect.signature(func).parameters)).values()  # ignores self for methods.
         ak = Struct.from_keys_values(ak.name, ak.default)
         ak = ak.update(kwargs)
@@ -259,8 +257,9 @@ class Experimental:
         for key, val in ak.items():
             if key != "args" and key != "kwargs":
                 flag = False
-                if val.default is val.empty:  # not passed argument.
-                    if exclude_args: flag = True
+                if val is inspect._empty:  # not passed argument.
+                    if exclude_args:
+                        flag = True
                     else:
                         val = None
                         print(f'Experimental Warning: arg {key} has no value. Now replaced with None.')
@@ -274,7 +273,7 @@ class Experimental:
             res += f"{ak.varkw} = " + "{}\n"
 
         if copy2clipboard:
-            clipboard = Experimental.assert_package_installed("clipboard")
+            clipboard = assert_package_installed("clipboard")
             clipboard.copy(res)
         if verbose: print("Finished. Paste code now.")
         return res
@@ -296,6 +295,7 @@ class Experimental:
                 raise KeyError(f"No marker found in the text. Place the following: 'here{line_idx}'")
         newsource = "\n".join(sourcelines)
         P(module.__file__).write_text(newsource)
+        import importlib
         importlib.reload(module)
         return module
 
@@ -321,7 +321,7 @@ class Experimental:
         else:
             raise KeyError(f"The pointer `{pointer}` was not found in the module `{module}`")
         print(cell)
-        clipboard = Experimental.assert_package_installed("clipboard")
+        clipboard = assert_package_installed("clipboard")
         clipboard.copy(cell)
         return cell
 
@@ -468,15 +468,64 @@ def batcherv2(func_type='function', order=1):
 
 
 class Terminal:
-    def __init__(self, stdout=None, stderr=None, elevated=False):
+    class Response:
+        @staticmethod
+        def from_completed_process(cp: subprocess.CompletedProcess):
+            tmp = dict(stdout=cp.stdout, stderr=cp.stderr, returncode=cp.returncode)
+            # for key, val in tmp:
+            #     if val is str:
+            #         tmp[key] = val.rstrip()
+            resp = Terminal.Response(cmd=cp.args)
+            resp.output.update(tmp)
+            return resp
+
+        def __init__(self, stdin=None, stdout=None, stderr=None, cmd=None):
+            self.std = dict(stdin=stdin, stdout=stdout, stderr=stderr)  # streams go here.
+            self.output = dict(stdin=None, stdout=None, stderr=None, returncode=None)
+            self.input = cmd  # input command
+
+        def __call__(self, *args, **kwargs):
+            return self.op.rstrip() if type(self.op) is str else None
+
+        @property
+        def op(self):
+            return self.output["stdout"]
+
+        @property
+        def ip(self):
+            return self.output["stdin"]
+
+        @property
+        def err(self):
+            return self.output["stderr"]
+
+        @property
+        def returncode(self):
+            return self.output["returncode"]
+
+        def capture(self):
+            for key, val in self.std.items():
+                if val is not None and val.readable():
+                    string = val.read().decode()
+                    self.output[key] = string.rstrip()  # move ending spaces and new lines.
+
+        def print(self):
+            self.capture()
+            print(f"Terminal Response:\nInput Command: {self.input}")
+            for idx, (key, val) in enumerate(self.output.items()):
+                print(f"{idx} - {key} {'=' * 30}\n{val}")
+            print("-" * 50, "\n\n")
+            return self
+
+    def __init__(self, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE, elevated=False):
         """
         Console
         Terminal
         Bash
         Shell
         Host
-        * adding console to the begining of the command results in launching a new console that will not
-        inherent from the console python was launched from (e.g. conda enviroment), unlike when console name is ignored.
+        * adding `start` to the begining of the command results in launching a new console that will not
+        inherit from the console python was launched from (e.g. conda enviroment), unlike when console name is ignored.
 
         * `subprocess.Popen` (process open) is the most general command. Used here to create asynchronous job.
         * `subprocess.run` is a thin wrapper around Popen that makes it wait until it finishes the task.
@@ -486,94 +535,238 @@ class Terminal:
         * To launch a new window, either use
         """
         self.available_consoles = ["cmd", "Command Prompt", "wt", "powershell", "wsl", "ubuntu", "pwsh"]
-        self.stdout = subprocess.DEVNULL if stdout is None else stdout
-        self.stderr = subprocess.DEVNULL if stderr is None else stderr
         self.elevated = elevated
-        # https://stackoverflow.com/questions/130763/request-uac-elevation-from-within-a-python-script
+        self.stdout = stdout
+        self.stderr = stderr
+        self.stdin = stdin
+        import platform
+        self.machine = platform.system()  # Windows, Linux, Darwin
+
+    def set_std_system(self):
+        self.stdout = sys.stdout
+        self.stderr = sys.stderr
+        self.stdin = sys.stdin
+
+    def set_std_pipe(self):
+        self.stdout = subprocess.PIPE
+        self.stderr = subprocess.PIPE
+        self.stdin = subprocess.PIPE
+
+    def set_std_null(self):
+        """Equivalent to `echo 'foo' &> /dev/null`"""
+        self.stdout = subprocess.DEVNULL
+        self.stderr = subprocess.DEVNULL
+        self.stdin = subprocess.DEVNULL
 
     @staticmethod
     def is_admin():
+        # https://stackoverflow.com/questions/130763/request-uac-elevation-from-within-a-python-script
         import ctypes
         return Experimental.try_this(lambda: ctypes.windll.shell32.IsUserAnAdmin(), otherwise=False)
 
-    def run_command(self, command, console="powershell"):
-        # alternative: res = subprocess.run("powershell -ls; dir", capture_output=True, shell=True, text=True)
-        my_list = [console, "-Command"] if console is not None else []
-        my_list.append(command)
+    def run(self, *cmds, shell=None, check=False, ip=None):
+        """Blocking operation. Thus, if you start a shell via this method, it will run in the main and
+        won't stop until you exit manually IF stdin is set to sys.stdin, otherwise it will run and close quickly.
+        Other combinations of stdin, stdout can lead to funny behaviour like no output but accept input or opposite.
+
+        * This method is short for:
+        res = subprocess.run("powershell command", capture_output=True, shell=True, text=True)
+        * Unlike `os.system(cmd)`, `subprocess.run(cmd)` gives much more control over the output and input.
+        * `shell=True` loads up the profile of the shell called so more specific commands can be run.
+            Importantly, on Windows, the `start` command becomes availalbe and new windows can be launched.
+        * `text=True` converts the bytes objects returned in stdout to text by default.
+
+        :param shell:
+        :param ip:
+        :param check: throw an exception is the execution of the external command failed (non zero returncode)
+        """
+        my_list = []
+        if self.machine == "Windows":
+            if shell in {"powershell", "pwsh"}:
+                my_list = [shell, "-Command"]  # alternatively, one can run "cmd"
+                """ The advantage of addig `powershell -Command` is to give access to wider range of options.
+            Other wise, command prompt shell doesn't recognize commands like `ls`."""
+            else:
+                pass  # that is, use command prompt as a shell, which is the default.
+        my_list += list(cmds)
         if self.elevated is False or self.is_admin():
-            resp = subprocess.run(my_list, capture_output=True, text=True, shell=True)
+            resp = subprocess.run(my_list, stderr=self.stderr, stdin=self.stdin, stdout=self.stdout,
+                                  text=True, shell=True, check=check, input=ip)
+            """
+            `capture_output` prevents the stdout to redirect to the stdout of the script automatically, instead it will
+        be stored in the Response object returned.
+            # `capture_output=True` same as `stdout=subprocess.PIPE, stderr=subprocess.PIPE`
+            """
         else:
             import ctypes
             resp = ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, " ".join(sys.argv), None, 1)
-        print(f"Crocodile: meta: Terminal: command execution response: {resp}")
-        return resp
+        return self.Response.from_completed_process(resp)
 
-    def run_command_async(self, command, console=None):
-        my_list = [console, "-Command"] if console is not None else []
-        my_list.append(command)
-        w = subprocess.Popen(my_list, stdout=self.stdout, stderr=self.stderr, shell=True)
+    def run_async(self, *cmds, new_window=True, shell=None, terminal=None):
+        """Opens a new terminal, and let it run asynchronously.
+        Maintaining an ongoing conversation with another process is very hard. It is adviseable to run all
+        commands in one go without interaction with an ongoing channel. Use this only for the purpose of
+        producing a different window and humanly interact with it.
+        https://stackoverflow.com/questions/54060274/dynamic-communication-between-main-and-subprocess-in-python
+        https://www.youtube.com/watch?v=IynV6Y80vws
+
+        https://www.oreilly.com/library/view/windows-powershell-cookbook/9781449359195/ch01.html
+        """
+        if terminal is None:
+            terminal = ""  # this means that cmd is the default console. alternative is "wt"
+        if shell is None:
+            if self.machine == "Windows":
+                shell = ""  # other options are "powershell" and "cmd"
+                # if terminal is wt, then it will pick powershell by default anyway.
+            else:
+                shell = ""
+        if new_window is True:  # start is alias for Start-Process which launches a new window.
+            new_window = "start"
+        else:
+            new_window = ""
+        if shell in {"powershell", "pwsh"}:
+            extra = "-Command"
+        else:
+            extra = ""
+        my_list = []
+        if self.machine == "Windows":
+            my_list += [new_window, terminal, shell, extra]
+        my_list += cmds
+        print("Meta.Terminal.run_async: Subprocess command: ", my_list)
+        my_list = [item for item in my_list if item != ""]
+        w = subprocess.Popen(my_list, stdin=subprocess.PIPE, shell=True)  # stdout=self.stdout, stderr=self.stderr, stdin=self.stdin
+        # returns Popen object, not so useful for communcation with an opened terminal
         return w
 
     @staticmethod
-    def open_console(console="", command="", shell=True, new_window=True, new_context=True):
-        """
-        :param console: default is same as the launching console. `cmd` doesn't recieve command, spoiler: not easy.
-         `powershell` doesn't inherit venv. `wt` and `` works.
-        :param command:
-        :param shell:
-        :param new_window:
-        :param new_context:
-        :return:
-        """
-        # This does not inherit from the from the shell launched python.
-        if new_context:
-            return subprocess.Popen(f'{"start" if new_window else ""} {console} {command}', shell=shell,
-                                    creationflags=subprocess.CREATE_NEW_CONSOLE)
-        else:  # this way, the new console inherits from its current context.
-            return os.system(fr'{"start" if new_window else ""} {console} {command} \K')
-            # /K remains the window, /C executes and dies (popup)
-
-    def run_script(self, script, wdir=None, interactive=True, shell=True, delete=False, console="", new_window=True):
-        """
+    def run_script(script, wdir=None, interactive=True, ipython=True,
+                   shell=None, delete=False, terminal="", new_window=True):
+        """This method is a wrapper on top of `run_async" except that the command passed will launch python
+        terminal that will run script passed by user.
         * Regular Python is much lighter than IPython. Consider using it while not debugging.
         """
         wdir = wdir or P.cwd()
         header = f"""
+# The following lines of code form a header appended by Terminal.run_script
 import crocodile.toolbox as tb
 tb.sys.path.insert(0, r'{wdir}')
+# End of header, start of script passed:
 """  # this header is necessary so import statements in the script passed are identified relevant to wdir.
 
         script = header + script
-        if console in {"wt", "powershell", "pwsh"}:
+        if terminal in {"wt", "powershell", "pwsh"}:
             script += "\ntb.DisplayData.set_pandas_auto_width()\n"
         script = f"""print(r'''{script}''')""" + "\n" + script
         file = P.tmpfile(name="tmp_python_script", suffix=".py", folder="tmpscripts")
         file.write_text(script)
-        print(f"Script to be executed asyncronously: ", file.as_uri())
-        self.open_console(console=console, command=f"ipython {'-i' if interactive else ''} {file}",
-                          shell=shell, new_window=new_window)
+        print(f"Script to be executed asyncronously: ", file.absolute().as_uri())
+        Terminal().run_async(f"{'ipython' if ipython else 'python'}",
+                             f"{'-i' if interactive else ''}",
+                             f"{file}",
+                             terminal=terminal, shell=shell, new_window=new_window)
         # python will use the same dir as the one from console this method is called.
-        # file.delete(are_you_sure=delete, verbose=False)
+        # file.delete(sure=delete, verbose=False)
         _ = delete
-        # TODO: add return option (asynchronous programming)
         # command = f'ipython {"-i" if interactive else ""} -c "{script}"'
 
-    def run_function(self, func):
-        """Python brachnes off to a new window and execute the function passed.
+    @staticmethod
+    def load_object_in_new_session(obj):
+        """Python brachnes off to a new window and run the function passed.
         context can be either a pickled session or the current file __file__"""
         # step 1: pickle the function
         # step 2: create a script that unpickles it.
         # step 3: run the script that runs the function.
         # TODO complete this
-        fname = P.tmpfile()
-        dill.dump(obj=func, file=fname)
+        fname = P.tmpfile(tstamp=False, suffix=".pkl")
+        Save.pickle(obj=obj, path=fname, verbose=False)
         script = f"""
-import crocodile.toolbox as tb
-func = dill.unpickle({fname})
-func()
+fname = tb.P(r'{fname}')
+obj = fname.readit()
+fname.delete(sure=True, verbose=False)
 """
-        # TODO: make sure that pickling function that serialize to tmp location delete this location afterwards
-        self.run_script(script)
+        Terminal.run_script(script)
+
+
+class SSH(object):
+    def __init__(self, hostname, username, ssh_key=None):
+        _ = False
+        if _:
+            super().__init__()
+        import paramiko
+        self.ssh_key = ssh_key
+        self.ssh = paramiko.SSHClient()
+        self.ssh.load_system_host_keys()
+        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
+        self.hostname = hostname
+        self.username = username
+        self.ssh.connect(hostname=hostname,
+                         username=username,
+                         port=22, key_filename=self.ssh_key.string if self.ssh_key is not None else None)
+        self.sftp = self.ssh.open_sftp()
+        self.load_python_cmd = rf"""source ~/miniconda3/bin/activate"""  # possible activate an env
+        import platform
+        self.platform = platform
+        self.target_machine = self.ssh.exec_command(self.load_python_cmd +
+                                                    "python -c 'import platform; platform.system()'")
+
+    def get_key(self):
+        """In SSH commands you need this:
+        scp -r {self.get_key()} "{str(source.expanduser())}" "{self.username}@{self.hostname}:'{target}'"
+        """
+        if self.ssh_key is not None:
+            return f"""-i "{str(P(self.ssh_key.expanduser()))}" """
+        else:
+            return ""
+
+    def __repr__(self):
+        return f"{self.local()} SSH connection to {self.remote()}"
+
+    def remote(self):
+        return f"{self.username}@{self.hostname}"
+
+    def local(self):
+        return f"{os.getlogin()}@{self.platform.node()}"
+
+    def open_console(self, new_window=True):
+        cmd = f"""ssh -i {self.ssh_key} {self.username}@{self.hostname}"""
+        Terminal().run_async(cmd, new_window=new_window)
+
+    def copy_env_var(self, name):
+        assert self.target_machine == "Linux"
+        self.run(f"{name} = {os.environ[name]}; export {name}")
+
+    def copy_from_here(self, source, target=None, zip_n_encrypt=True):
+        pwd = None
+        if zip_n_encrypt:
+            pwd = randstr(length=10, safe=True)
+            source = P(source).expanduser().zip_n_encrypt(pwd=pwd)
+        print(f"Step 1: Creating Target directory {target} @ remote machine.")
+        if target is None: target = P(source).collapseuser().as_posix()  # works if source is relative.
+        resp = self.runpy(f'print(tb.P(r"{target}").expanduser().parent.create())')
+        remotepath = P(resp.op or "").joinpath(P(target).name).as_posix()
+        print(f"Send `{source}` ==> `{remotepath}` ")
+        self.sftp.put(localpath=P(source).expanduser(), remotepath=remotepath)
+        if zip_n_encrypt:
+            resp = self.runpy(f"""tb.P(r"{remotepath}").expanduser().decrypt_n_unzip(pwd="{pwd}", delete=True)""")
+            return resp
+
+    def copy_to_here(self, source, target=None):
+        pass
+
+    def run(self, cmd, printit=True):
+        print(f"\nExecuting on remote {self.username}@{self.hostname}:\n{cmd}")
+        res = self.ssh.exec_command(cmd)
+        res = Terminal.Response(stdin=res[0], stdout=res[1], stderr=res[2], cmd=cmd)
+        if printit: res.print()
+        return res
+
+    def runpy(self, cmd):
+        cmd = f"""{self.load_python_cmd}; python -c 'import crocodile.toolbox as tb; {cmd} ' """
+        return self.run(cmd)
+
+    def run_locally(self, command):
+        print(f"Executing Locally @ {self.platform.node()}:\n{command}")
+        return Terminal.Response(os.system(command))
 
 
 class Log(object):
@@ -628,6 +821,10 @@ class Log(object):
     def critical(self, msg):
         return self.logger.critical(msg)
 
+    @property
+    def file(self):
+        return P(self.specs["file_path"]) if self.specs["file_path"] else None
+
     def _install(self):  # populates self.logger attribute according to specs and dielect.
         if self.dialect == "colorlog":
             self.logger = Log.get_colorlog(log_colors=self.log_colors, **self.specs)
@@ -690,9 +887,9 @@ class Log(object):
                         'name': {'color': 'blue'},
                         'programname': {'color': 'cyan'},
                         'username': {'color': 'yellow'}}
-        coloredlogs = Experimental.assert_package_installed("coloredlogs")
+        coloredlogs = assert_package_installed("coloredlogs")
         if verbose:
-            verboselogs = Experimental.assert_package_installed("verboselogs")
+            verboselogs = assert_package_installed("verboselogs")
             # https://github.com/xolox/python-verboselogs
             # verboselogs.install()  # hooks into logging module.
             logger = verboselogs.VerboseLogger(name=name)
@@ -719,7 +916,7 @@ class Log(object):
                           'ERROR': 'thin_red',
                           'CRITICAL': 'fg_bold_red,bg_white',
                           }  # see here for format: https://pypi.org/project/colorlog/
-        colorlog = Experimental.assert_package_installed("colorlog")
+        colorlog = assert_package_installed("colorlog")
         logger = Log.get_base_logger(colorlog, name, l_level)
         fmt = colorlog.ColoredFormatter(fmt or (rf"%(log_color)s" + Log.get_format(sep)), log_colors=log_colors)
         Log.add_handlers(logger, colorlog, file, f_level, file_path, fmt, stream, s_level)
@@ -768,7 +965,8 @@ class Log(object):
         fhandler.setLevel(level=f_level)
         fhandler.set_name(name)
         logger.addHandler(fhandler)
-        print(f"    Level {f_level} file handler for Logger `{logger.name}` is created @ " + P(file_path).as_uri())
+        print(f"    Level {f_level} file handler for Logger `{logger.name}` is created @ " + P(
+            file_path).absolute().as_uri())
 
     @staticmethod
     def test_logger(logger):
@@ -840,10 +1038,10 @@ class Scheduler:
         self.occasional = occasional  # routine to be repeated every `other` time.
         self.exception_handler = exception if exception is not None else lambda ex: None
         self.wind_down = wind_down
-        # routine to be run_command when an error occurs, e.g. save object.
+        # routine to be run when an error occurs, e.g. save object.
         self.wait = wait  # wait period between routine cycles.
         self.other = other  # number of routine cycles before `occasional` get executed once.
-        self.cycles = runs  # how many times to run_command the routine. defaults to infinite.
+        self.cycles = runs  # how many times to run the routine. defaults to infinite.
         self.logger = logger or Log(name="SchedulerAutoLogger" + randstr())
         self.history = []
         self._start_time = None  # begining of a session (local time)
@@ -855,6 +1053,7 @@ class Scheduler:
         self.count = 0
         self._start_time = datetime.now()
         wait_time = str2timedelta(self.wait).total_seconds()
+        import pandas as pd
         until = pd.to_datetime(until)  # (local time)
 
         while datetime.now() < until and self.count < self.cycles:
@@ -922,6 +1121,15 @@ class Scheduler:
         # signal.signal(signal.SIGINT, keyboard_interrupt_handler)
 
 
+def qr(txt):
+    qrcode = assert_package_installed("qrcode")
+    img = qrcode.make(txt)
+    file = P.tmpfile(suffix=".png")
+    img.save(file.__str__())
+    file()
+    return file
+
+
 if __name__ == '__main__':
     # Log.get_colorlog()
-    pass
+    Terminal().run("$profile", shell="pwsh").print()
